#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//************************************************************************
// This program reads two Spin binary files and links them together.
// The first file is a parent binary object file.  It must contain a
// stu object, which will be replaced by the child binary.  The methods
// defined in the stub object must match the methods in the top object
// of the child binary, such that the order of the methods is the same
// and the number of parameters in the methods must be the same.
//
// The second file is the child binary file.  It may contain more than
// one object, but the top object must match the stub object in the
// parent as described above.
//
// This program will replace the stub object in the parent with the
// objects in the child binary file.  The object number of the stub
// object must be specified unless it is the last object in the parent
// binary file.  The objects are numbered starting with zero for the top
// object.
//
// After replacing the stub object with the child, the object addresses
// and the VAR offset will be adjusted to match the new objects.  The
// 16-byte program header is also adjusted.
//
// Dave Hein, Feburary 2, 2011
//************************************************************************

// Define a 64K buffer
static unsigned short buffer[32768];

int printflag = 0;

void usage(void)
{
    printf("usage: linkit parent.binary [-n] child.binary out.binary\n");
    printf("  -n is an optional parameter that specifie the object\n");
    printf("  number to be replaced by the child.  The top object is\n");
    printf("  numbered 0, and cannot be specified.  The default is the\n");
    printf("  last object in the parent.\n");
    exit(1);
}

FILE *OpenFile(char *fname, char *mode)
{
    FILE *stream = fopen(fname, mode);

    if (!stream)
    {
	printf("Could not open %s\n", fname);
	exit(1);
    }

    return stream;
}

static int objnum = -1;
static int numobjs = 0;
static int objsizes[100];
static int oldobjaddr[100];
static int newobjaddr[100];
static int objvarsize[100];
static int objvartotal[100];

int MapObjAddrToIndex(int addr)
{
    int i;

    for (i = 0; i < numobjs; i++)
    {
	if (addr == oldobjaddr[i]) return i;
    }
    printf("MapObjAddrToIndex: Object address %4.4x not found\n", addr);
    exit(1);
    return 0;
}

int MapNewObjAddrToIndex(int addr)
{
    int i;

    for (i = 0; i < numobjs; i++)
    {
	if (addr == newobjaddr[i]) return i;
    }
    printf("MapNewObjAddrToIndex: Object address %4.4x not found\n", addr);
    exit(1);
    return 0;
}

int AdjustAddress(int oldaddr)
{
    int i = MapObjAddrToIndex(oldaddr);

    if (oldaddr == newobjaddr[i]) return oldaddr;
    if (printflag)
        printf("AdjustAddress: %4.4x %4.4x\n", oldaddr, newobjaddr[i]);
    return newobjaddr[i];
}

void SetObjectVarTotal(int objaddr, int vartotal)
{
    int i = MapObjAddrToIndex(objaddr);

    if (objaddr == oldobjaddr[i])
    {
	if (printflag)
	{
            printf("SetObjectVarTotal: addr = %4.4x, total = %4.4x\n",
                objaddr, vartotal);
            if (objvartotal[i] != -1 && objvartotal[i] != vartotal)
                printf("SetObjectVartotal: Changing from %4.4x to %4.4x\n",
                    objvartotal[i], vartotal);
	}
        objvartotal[i] = vartotal;
        return;
    }

    printf("SetObjectVarTotal: Could find address %4.4x\n", objaddr);
}

void SetObjectVarSize(int objaddr, int varsize)
{
    int i = MapObjAddrToIndex(objaddr);

    if (objaddr == oldobjaddr[i])
    {
	if (printflag)
	{
            printf("SetObjectVarSize: addr = %4.4x, size = %4.4x\n",
                objaddr, varsize);
            if (objvarsize[i] != -1 && objvarsize[i] != varsize)
                printf("SetObjectVarSize: Changing from %4.4x to %4.4x\n",
                    objvarsize[i], varsize);
	}
        objvarsize[i] = varsize;
        return;
    }

    if (printflag)
        printf("SetObjectVarSize: Could find address %4.4x\n", objaddr);
}

void TraverseMethodPath(int addr0, int varsize0)
{
    int addr;
    int i, j;
    int index;
    int numm, numo;
    int lastval, lastaddr;

    addr = addr0;
    numm = buffer[(addr/2)+1];
    numo = numm >> 8;
    numm = (numm & 255) - 1;
    if (numo == 0)
    {
	SetObjectVarSize(addr, varsize0);
	return;
    }

    if (printflag)
        printf("TraverseMethodPath: %4.4x %4.4x %d %d\n", addr, varsize0, numm, numo);

    // Set this object's VAR size and the objects in the method table
    index = (addr/2) + (numm*2) + 2;
    lastval = 0;
    lastaddr = addr;
    for (j = 0; j < numo; j++)
    {
        if (printflag)
	    printf("%d: %4.4x %4.4x\n", j, buffer[index]+addr, buffer[index+1]);
	if (j == 0)
            SetObjectVarSize(lastaddr, buffer[index+1] - lastval);
	else
            SetObjectVarTotal(lastaddr, buffer[index+1] - lastval);
        lastaddr = buffer[index] + addr;
        lastval = buffer[index+1];
        index += 2;
    }
    SetObjectVarTotal(lastaddr, varsize0 - lastval);

    // Now check the referenced objects for more objects
    index = (addr0/2) + (numm*2) + 2;
    for (j = 0; j < numo; j++)
    {
	addr = buffer[index] + addr0;
	i = MapObjAddrToIndex(addr);
	TraverseMethodPath(addr, objvartotal[i]);
        index += 2;
    }
}

int ComputeTotalVarSize(int addr0)
{
    int i;
    int addr;
    int index;
    int varsize;
    int numm, numo;

    i = MapObjAddrToIndex(addr0);
    varsize = objvarsize[i];
    index = (addr0/2) + 1;
    numm = buffer[index];
    numo = numm >> 8;
    numm = (numm & 255) - 1;
    index += (numm*2) + 1;

    for (i = 0; i < numo; i++)
    {
	addr = buffer[index] + addr0;
	varsize += ComputeTotalVarSize(addr);
        index += 2;
    }

    SetObjectVarTotal(addr0, varsize);

    return varsize;
}

void UpdateVarOffset(int addr0)
{
    int addr;
    int i, j;
    int index;
    int offset;
    int numm, numo;

    return;

    numm = buffer[(addr0/2)+1];
    numo = numm >> 8;
    numm = (numm & 255) - 1;
    if (numo == 0) return;

    if (printflag)
        printf("UpdateVarOffset: %4.4x %d %d\n", addr0, numm, numo);

    i = MapNewObjAddrToIndex(addr0);
    if (i == objnum) return;
    offset = objvarsize[i];

    // Update the VAR offsets
    index = (addr0/2) + (numm*2) + 2;
    for (j = 0; j < numo; j++)
    {
        if (printflag)
	    printf("UpdateVarOffset: change from %4.4x %4.4x\n", buffer[index+1], offset);
	buffer[index+1] = offset;
	addr = buffer[index] + addr0;
	UpdateVarOffset(addr);
	i = MapNewObjAddrToIndex(addr);
	offset += objvartotal[i];
	index += 2;
    }
}

void main(int argc, char **argv)
{
    int i, j;
    int addr;
    int numm, numo;
    FILE *infile;
    FILE *infile2;
    FILE *outfile;
    int stackoff;
    int varsize1, varsize2;
    int objsize1, objsize2;
    int checksum = 0;
    int filesize;
    int objsize;
    unsigned char *byteptr = (unsigned char *)buffer;
    char *parent;
    char *child;
    char *output;
    int k;
    int index;

    // Check for print flag
    if (strcmp(argv[argc-1], "-p") == 0)
    {
	printflag = 1;
	argc--;
    }

    // Get command-line parameters
    parent = argv[1];
    if (argc == 5)
    {
	if (argv[2][0] != '-' || argv[2][1] < '0' || argv[2][1] > '9') usage();
	sscanf(&argv[2][1], "%d", &objnum);
	if (objnum <= 0) usage();
	child = argv[3];
	output = argv[4];
    }
    else
    {
	if (argc != 4) usage();
	child = argv[2];
	output = argv[3];
    }

    // Read the parent file and print the object information
    infile = OpenFile(parent, "rb");
    filesize = fread(buffer, 1, 32768, infile);
    fclose(infile);

    if (printflag)
    {
        printf("Parent object %s\n", parent);
        printf("pbase = %4.4x, vbase = %4.4x, dbase = %4.4x, pcurr = %4.4x, dcurr = %4.4x\n",
            buffer[3], buffer[4], buffer[5], buffer[6], buffer[7]);
        printf("Object size = %4.4x, ", buffer[4]);
        printf("VAR size = %4.4x\n", buffer[5] - buffer[4] - 8);
        printf("\n");
    }

    // Extract the VAR size and stack offset
    varsize1 = buffer[5] - buffer[4] - 8;
    stackoff = buffer[7] - buffer[5];

    // Count the number of objects and save the size and address
    if (printflag) printf("Parent object addresses\n");
    numobjs = 0;
    objsize1 = 0;
    while (1)
    {
	objsize = buffer[(16+objsize1)/2];
	objsizes[numobjs] = objsize;
	oldobjaddr[numobjs] = 16 + objsize1;
	objvarsize[numobjs] = -1;
	objvartotal[numobjs] = -1;
	if (printflag)
	    printf("obj %d: %4.4x\n", numobjs, oldobjaddr[numobjs]);
	numobjs++;
	if (objsize1 + objsize >= filesize - 16) break;
	objsize1 += objsize;
    }
    if (printflag) printf("\n");

    if (objnum < 0)
    {
	objnum = numobjs - 1;
    }
    else if (numobjs - 1 < objnum)
    {
	printf("Specified object number is too large\n");
	exit(1);
    }

    // Read the child header and print the object information
    infile = OpenFile(child, "rb");
    filesize = fread(buffer, 1, 16, infile);
    fclose(infile);

    if (printflag)
    {
        printf("Child object %s\n", child);
        printf("pbase = %4.4x, vbase = %4.4x, dbase = %4.4x, pcurr = %4.4x, dcurr = %4.4x\n",
            buffer[3], buffer[4], buffer[5], buffer[6], buffer[7]);
        printf("Object size = %4.4x, ", buffer[4]);
        printf("VAR size = %4.4x\n", buffer[5] - buffer[4] - 8);
        printf("\n");
    }

    // Extract the object size and VAR size
    objsize2 = buffer[4];
    varsize2 = buffer[5] - buffer[4] - 8;

    // Recursively traverse the object path and determine the VAR sizes
    if (printflag) printf("Determine the VAR sizes in the parent\n");
    SetObjectVarTotal(16, varsize1);
    TraverseMethodPath(16, varsize1);
    if (printflag) printf("\n");

    // Print the object information
    if (printflag)
    {
        printf("Object info\n");
        for (i = 0; i < numobjs; i++)
        {
	    printf("%d: %4.4x %4.4x %4.4x %4.4x\n",
	        i, oldobjaddr[i], objsizes[i], objvarsize[i], objvartotal[i]);
        }
        printf("\n");
    }

    // Change the VAR size for the replacement object
    if (printflag)
    {
        printf("Changing VAR size for object %d from %4.4x to %4.4x\n",
            objnum, objvarsize[objnum], varsize2);
        printf("Recalculating the total VAR size for each object\n");
    }
    objvarsize[objnum] = varsize2;
    for (i = 0; i < numobjs; i++) objvartotal[i] = -1;
    ComputeTotalVarSize(16);
    if (printflag) printf("\n");

    // Print the new object information
    if (printflag)
    {
        printf("New object info\n");
        for (i = 0; i < numobjs; i++)
        {
	    printf("%d: %4.4x %4.4x %4.4x %4.4x\n",
	        i, oldobjaddr[i], objsizes[i], objvarsize[i], objvartotal[i]);
        }
        printf("\n");
    }

    // Open the parent file and read the header
    byteptr = (unsigned char *)buffer;
    infile = OpenFile(parent, "rb");
    fread(buffer, 1, 16, infile);
    byteptr += 16;

    // Open the child file and skip the header
    infile2 = OpenFile(child, "rb");
    fread(byteptr, 1, 16, infile2);

    // Read the parent's objects and add the child where specified
    for (i = 0; i < numobjs; i++)
    {
	objsize = objsizes[i];
	newobjaddr[i] = ((int)byteptr) - ((int)buffer);
	if (printflag)
	{
	    printf("obj %d: %4.4x %4.4x %4.4x\n",
	        i, oldobjaddr[i], newobjaddr[i], objvarsize[i]);
	}
        fread(byteptr, 1, objsize, infile);
	if (i != objnum)
	{
	    if (printflag) printf("Add parent object\n");
	    byteptr += objsize;
	}
	else
	{
	    if (printflag) printf("Add child objects\n");
            fread(byteptr, 1, objsize2 - 16, infile2);
	    byteptr += objsize2 - 16;
	}
    }
    newobjaddr[i] = 1000000000;
    fclose(infile);
    fclose(infile2);

    // Print the new object information with the new addresses
    if (printflag)
    {
        printf("New object info with new addresses\n");
        for (i = 0; i < numobjs; i++)
        {
	    printf("%d: %4.4x %4.4x %4.4x %4.4x %4.4x\n",
	        i, oldobjaddr[i], newobjaddr[i], objsizes[i], objvarsize[i], objvartotal[i]);
        }
        printf("\n");
    }

    // Set VBASE in the header to the sum of the object size
    // plus the header size
    buffer[4] = ((int)byteptr) - ((int)buffer);

    // Set DBASE in the header to the VBASE plus the sum of
    // the VAR sizes plus 8 for the top return frame
    buffer[5] = buffer[4] + varsize1 + varsize2 + 8;

    // Set DCURR in the header to DBASE plus the stack offset.
    // The stack offset is normally 4, but it could be larger if
    // the first method contains calling parameters.
    buffer[7] = buffer[5] + stackoff;

    // Fix the object pointers in the method tables
    i = 0;
    k = 0;
    addr = 16;
    if (printflag) printf("buffer[8] = %4.4x\n", buffer[8]);
    while (addr < buffer[4])
    {
	if (printflag) printf("addr = %4.4x, buffer[4] = %4.4x\n", addr, buffer[4]);
	objsize = buffer[addr/2];
	numm = buffer[(addr/2)+1];
	numo = numm >> 8;
	numm = (numm & 255) - 1;
	if (addr != newobjaddr[i])
	{
	    if (printflag) printf("%d: %4.4x %4.4x %d %d %d\n", k, addr, objsize, numm, numo, 0);
	    k++;
	    addr += objsize;
	    continue;
	}
	if (printflag) printf("%d: %4.4x %4.4x %d %d %d\n", k, addr, objsize, numm, numo, 1);
	index = (addr/2) + (numm*2) + 2;
	for (j = 0; j < numo; j++)
	{
	    if (printflag) printf("%d: %4.4x %4.4x\n", j, buffer[index] + addr, buffer[index + 1]);
	    buffer[index] = AdjustAddress(buffer[index] + addr) - addr;
	    index += 2;
	}
	i++;
	k++;
	addr += objsize;
    }

    // Update the VAR offsets
    if (printflag) printf("Update the object VAR size entries\n");
    UpdateVarOffset(16);

    // Compute the new checksum and write it at header byte 5
    byteptr = (unsigned char *)buffer;
    for (i = 0; i < buffer[5]; i++) checksum += byteptr[i];
    byteptr[5] += 0x14 - checksum;

    if (printflag) printf("\n");

    // Print the new program values
    if (printflag)
    {
        printf("Linked program %s\n", output);
        printf("pbase = %4.4x, vbase = %4.4x, dbase = %4.4x, pcurr = %4.4x, dcurr = %4.4x\n",
            buffer[3], buffer[4], buffer[5], buffer[6], buffer[7]);
        printf("Object size = %4.4x, ", buffer[4]);
        printf("VAR size = %4.4x\n", buffer[5] - buffer[4] - 8);
    }

    // Write the output file
    outfile = OpenFile(output, "wb");
    fwrite(buffer, 1, buffer[4], outfile);
    fclose(outfile);
}
